;
; sat68kdb: Sega Saturn MC68EC000 Debugger
; Bart Trzynadlowski
;

;
; driv_68k.asm: Sega Saturn 68000 driver code. Assemble with SNASM68K. The
; "org" directive should not affect file size. Assumes memory is clear when
; it starts up.
;

;
; Commands (not including commands processed by SH-2):
; ----------------------------------------------------
; 0x0000    No command (68K continues waiting for command.)
; 0x0001    Execute next instruction (trace.)
; 0x0002    Read PC. Result is written to MEM_DATA.
; 0x0003    Read register. MEM_DATA contains register offset and the result
;           when the operation is finished.
; 0x0004    Write register. MEM_DATA contains register offset, MEM_DATA2 has
;           the value to write.
; 0x0005    Read SR. MEM_DATA contains result.
; 0x0006    Write SR. MEM_DATA contains data to write.
; 0x0007    Write PC. MEM_DATA contains data to write.
; 0x0008    Read XSP (whichever SP A7 is not.) MEM_DATA contains result value. 
; 

;
; Memory Definitions
;

STACK_TOP       equ $7fe00  ; top of stack

MEM_STARTSIG    equ $7fff2  ; if 0, do not start, if 1, a program has been
                            ; loaded and execution may commence
MEM_READY       equ $7fff4  ; if 1, awaiting a command (results of last
                            ; command are ready), if 0, processing a command.
                            ; 0 must be written by the SH-2 when a command
                            ; is written
MEM_COMMAND     equ $7fff6  ; command (WORD)
MEM_DATA        equ $7fff8  ; long word data buffer (LONG WORD)
MEM_DATA2       equ $7fffc  ; long word data buffer #2 (LONG WORD)


    org 0

;
; 68000 Vector Table
;
; Only partially filled up, it is assumed the rest will not ever have to be
; used.
;
    dc.l    STACK_TOP, Start    ; SP, PC
    dc.l    0                   ; bus error
    dc.l    0                   ; address error
    dc.l    0                   ; illegal instruction
    dc.l    0                   ; division by zero
    dc.l    0                   ; CHK exception
    dc.l    0                   ; TRAP exception
    dc.l    0                   ; privilege violation
    dc.l    $7fe00              ; trace


    org $7fe00
;
; All the work happens here
;
TraceHandler:
    ;
    ; Wait for a command
    ;    
    move.w  #1,MEM_READY        ; 68K is awaiting a command
wait_for_command:
    tst.w   MEM_READY           ; wait until SH-2 turns this to 0
    bne     wait_for_command

    ;
    ; Check to see which command it is and process it accordingly
    ;
    cmp.w   #$0001,MEM_COMMAND
    beq     _Trace
    cmp.w   #$0002,MEM_COMMAND
    beq     _ReadPC
    cmp.w   #$0003,MEM_COMMAND
    beq     _ReadRegister
    cmp.w   #$0004,MEM_COMMAND
    beq     _WriteRegister
    cmp.w   #$0005,MEM_COMMAND
    beq     _ReadSR
    cmp.w   #$0006,MEM_COMMAND
    beq     _WriteSR
    cmp.w   #$0007,MEM_COMMAND
    beq     _WritePC
    cmp.w   #$0008,MEM_COMMAND
    beq     _ReadXSP

    bra     TraceHandler        ; unknown, let's just say we did it

_Trace:
    rte                         ; go on to next instruction

_ReadPC:
    move.l  2(a7),MEM_DATA      ; RTE data is on the stack now
    bra     TraceHandler

_WritePC:
    move.l  MEM_DATA,2(a7)
    bra     TraceHandler

_ReadRegister:
    movem.l d0-d7/a0-a7,-(a7)   ; save all registers to stack
    ;
    ; If user wants A7, we have to do some fancy footwork to make sure the
    ; proper SP is returned
    ;
    cmp.w   #$3c,(MEM_DATA+2)
    bne     continue_readreg

    btst    #5,$40(a7)          ; see if S bit is set
    bne     continue_readreg    ; yes, no fancy trickery required

    move    usp,a0              ; if user mode, get the USP
    move.l  a0,MEM_DATA
    bra     end_readreg

continue_readreg:
    add.l   #6,$3c(a7)          ; adjust A7
    move.w  (MEM_DATA+2),d0     ; get register offset
    move.l  0(a7,d0.w),MEM_DATA ; index into stack, and write to MEM_DATA
    sub.l   #6,$3c(a7)          ; restore A7
end_readreg:
    movem.l (a7)+,d0-d7/a0-a7   ; restore all registers
    bra     TraceHandler

_WriteRegister:
    movem.l d0-d7/a0-a7,-(a7)   ; save all registers to stack
    ;
    ; If user wants A7, we have to do some fancy footwork to make sure the
    ; proper SP is handled
    ;
    cmp.w   #$3c,(MEM_DATA+2)
    bne     continue_writereg

    btst    #5,$40(a7)          ; see if S bit is set
    bne     continue_writereg   ; yes, no fancy trickery required

    move.l  MEM_DATA2,a0        ; if user mode, write to USP
    move    a0,usp
    move    usp,a0              ; if user mode, get the USP
    bra     end_writereg

continue_writereg:

    add.l   #6,$3c(a7)          ; adjust A7
    move.w  (MEM_DATA+2),d0     ; get register offset
    move.l  MEM_DATA2,0(a7,d0.w); index into stack, and write data to reg
    sub.l   #6,$3c(a7)          ; restore A7
end_writereg:
    movem.l (a7)+,d0-d7/a0-a7   ; restore all registers
    bra     TraceHandler

_ReadSR:
    move.w  (a7),(MEM_DATA+2)   ; RTE data is on the stack now
    bra     TraceHandler

_WriteSR:
    move.w  (MEM_DATA+2),(a7)
    and.w   #$a71f,(a7)         ; the real 68K does this, too
    bra     TraceHandler

_ReadXSP:
    btst    #5,(a7)             ; see if S bit is set
    bne     read_usp            ; yes, go read the USP

    move.l  a7,MEM_DATA         ; otherwise, A7 is SSP
    add.l   #6,MEM_DATA         ; adjust A7
    bra     TraceHandler

read_usp:
    move.l  a0,-(a7)
    move    usp,a0
    move.l  a0,MEM_DATA
    move.l  (a7)+,a0
    bra     TraceHandler


;
; Entry point
;
Start:
    clr.w   MEM_READY           ; busy
    movem.l $100.w,d0-d7/a0-a6  ; all memory is presumed to be clear, so we
                                ; can safely clear the registers by reading
                                ; from an arbitrary location
    move    a0,usp              ; clear USP as well
wait_to_start:
    tst.w   MEM_STARTSIG        ; wait until we're cleared to begin
    beq     wait_to_start
    move.l  $4,a0               ; start vector
    move    #$a700,sr           ; trace on any instruction
    jmp     (a0)                ; jump to the starting point
