/*
 * sat68kdb: Sega Saturn MC68EC000 Debugger
 * Bart Trzynadlowski
 */

/*
 * satintrf.c: Sega Saturn interface
 */
 
#include <pc.h>
#include "satintrf.h"
#include "satcode.h"

/*
 * Base Communications Port
 */

UINT16  par_base_port;

/*
 * Statistics
 */

UINT32  sat68kdb_bytes_transferred;


/*
 * PAR Communication
 */

UINT8 PARTransferByte(UINT8 data)
{
    outportb(par_base_port, data);
    sat68kdb_bytes_transferred++;
    while (inportb(par_base_port + 2) & 1)  ;
    sat68kdb_bytes_transferred++;
    return inportb(par_base_port);
}

void PARSendWord(UINT16 data)
{
    PARTransferByte(data >> 8);
    PARTransferByte(data & 0xff);
}

void PARSendLong(UINT32 data)
{
    PARTransferByte(data >> 24);
    PARTransferByte((data >> 16) & 0xff);
    PARTransferByte((data >> 8) & 0xff);
    PARTransferByte(data & 0xff);
}

UINT32 PARReceiveLong()
{
    UINT32  data;

    data = PARTransferByte(0) << 24;
    data |= PARTransferByte(0) << 16;
    data |= PARTransferByte(0) << 8;
    data |= PARTransferByte(0);

    return data;
}


/*
 * Saturn Communication
 */

/*
 * UINT32 sat68kdb_Command(UINT16 command, UINT32 data, UINT32 data2);
 *
 * Sends a command and data to the Saturn, returns the resultant data. If a
 * command does not return data, the return value of this function will be
 * undefined.
 */

UINT32 sat68kdb_Command(UINT16 command, UINT32 data, UINT32 data2)
{
    PARSendWord(command);       /* send the command */
    PARSendLong(data);          /* and the data ... */
    PARSendLong(data2);
    return PARReceiveLong();    /* return the data from the Saturn */
}

/*
 * void sat68kdb_Init(UINT16 port, UINT8 *prog, UINT32 size);
 *
 * Initializes communication with the Saturn and sends over an initial program
 * to load on the 68K side. The base communication port number is also
 * required for communication with the PAR. If the port number is 0, 0x300 is
 * assumed.
 */

void sat68kdb_Init(UINT16 port, UINT8 *prog, UINT32 size)
{
    UINT32  i;
    UINT16  j;

    /*
     * Set port
     */

    par_base_port = (!port) ? 0x300 : port;

    /*
     * Clear statistical data
     */

    sat68kdb_bytes_transferred = 0;

    /*
     * Handshake with PAR, and send our SH-2 driver
     */

    while (1)
    {
        if (PARTransferByte('I') != 'D')
            continue;
        if (PARTransferByte('N') == 'O')
            break;
    }

    PARTransferByte(9);
    PARSendLong(0x6004000);     /* start address of program */
    PARSendLong(SATCODE_SIZE);  /* size */
    PARTransferByte(1);

    for (i = 0; i < SATCODE_SIZE; i++)
        PARTransferByte(satcode[i]);
        
    /*
     * Handshake with our SH-2 driver
     */

    while (1)
    {
        if (PARTransferByte('M') != '6')
            continue;
        if (PARTransferByte('8') == 'K')
            break;
    }
   
    /*
     * Send the size of the initial program
     */

    PARSendLong(size);

    /*
     * Send the initial program, word by word
     */

    i = 0;
    while (size)
    {
        j = *((UINT16 *) &prog[i]);
        j = WORDSWAP(j);
        PARSendWord(j);
        i += 2;
        size -= 2;
    }
}

/*
 * void sat68kdb_Deinit();
 *
 * Resets the Saturn. We can't use sat68kdb_Command() because we never get a
 * response back from the Saturn.
 */

void sat68kdb_Deinit()
{
    PARSendWord(CMND_RESET);
    PARSendLong(0);
    PARSendLong(0);
}
